"""
This type stub file was generated by pyright.
"""

from ._common import *

def remove(src, new_file=...): # -> None:
    """
    py:function:: piexif.remove(filename)

    Remove exif from JPEG.

    :param str filename: JPEG
    """
    ...



"""
Pytest configuration and shared fixtures.

Provides cross-platform test file paths that work on both Windows and Linux (CI).
"""

from pathlib import Path

import pytest

# Get the tests directory (this file's parent)
TESTS_DIR = Path(__file__).parent
ASSETS_DIR = TESTS_DIR / "assets"
TEST_IMAGES_DIR = ASSETS_DIR / "test_images"


@pytest.fixture
def jpg_test_file() -> Path:
    """Return path to a JPG test file."""
    return TEST_IMAGES_DIR / "test_fuji.jpg"


@pytest.fixture
def png_test_file() -> Path:
    """Return path to a PNG test file."""
    return TEST_IMAGES_DIR / "generated_test_03.png"


@pytest.fixture
def test_images_dir() -> Path:
    """Return path to test images directory."""
    return TEST_IMAGES_DIR


# String versions for parametrize (which doesn't support fixtures directly)
def get_jpg_test_file() -> str:
    """Get JPG test file path as string."""
    return str(TEST_IMAGES_DIR / "test_fuji.jpg")


def get_png_test_file() -> str:
    """Get PNG test file path as string."""
    return str(TEST_IMAGES_DIR / "generated_test_03.png")


def get_test_images_dir() -> str:
    """Get test images directory path as string."""
    return str(TEST_IMAGES_DIR)



"""
Type stubs for piexif EXIF tag definitions.
"""

from typing import Final

class TYPES:
    """EXIF data type identifiers."""

    Byte: Final[int] = ...
    Ascii: Final[int] = ...
    Short: Final[int] = ...
    Long: Final[int] = ...
    Rational: Final[int] = ...
    SByte: Final[int] = ...
    Undefined: Final[int] = ...
    SShort: Final[int] = ...
    SLong: Final[int] = ...
    SRational: Final[int] = ...
    Float: Final[int] = ...
    DFloat: Final[int] = ...

TAGS: dict[str, dict[int, dict[str, str | int]]]

class ImageIFD:
    """Exif tag number reference - 0th IFD"""

    ProcessingSoftware: Final[int] = ...
    NewSubfileType: Final[int] = ...
    SubfileType: Final[int] = ...
    ImageWidth: Final[int] = ...
    ImageLength: Final[int] = ...
    BitsPerSample: Final[int] = ...
    Compression: Final[int] = ...
    PhotometricInterpretation: Final[int] = ...
    Threshholding: Final[int] = ...
    CellWidth: Final[int] = ...
    CellLength: Final[int] = ...
    FillOrder: Final[int] = ...
    DocumentName: Final[int] = ...
    ImageDescription: Final[int] = ...
    Make: Final[int] = ...
    Model: Final[int] = ...
    StripOffsets: Final[int] = ...
    Orientation: Final[int] = ...
    SamplesPerPixel: Final[int] = ...
    RowsPerStrip: Final[int] = ...
    StripByteCounts: Final[int] = ...
    XResolution: Final[int] = ...
    YResolution: Final[int] = ...
    PlanarConfiguration: Final[int] = ...
    GrayResponseUnit: Final[int] = ...
    GrayResponseCurve: Final[int] = ...
    T4Options: Final[int] = ...
    T6Options: Final[int] = ...
    ResolutionUnit: Final[int] = ...
    TransferFunction: Final[int] = ...
    Software: Final[int] = ...
    DateTime: Final[int] = ...
    Artist: Final[int] = ...
    HostComputer: Final[int] = ...
    Predictor: Final[int] = ...
    WhitePoint: Final[int] = ...
    PrimaryChromaticities: Final[int] = ...
    ColorMap: Final[int] = ...
    HalftoneHints: Final[int] = ...
    TileWidth: Final[int] = ...
    TileLength: Final[int] = ...
    TileOffsets: Final[int] = ...
    TileByteCounts: Final[int] = ...
    SubIFDs: Final[int] = ...
    InkSet: Final[int] = ...
    InkNames: Final[int] = ...
    NumberOfInks: Final[int] = ...
    DotRange: Final[int] = ...
    TargetPrinter: Final[int] = ...
    ExtraSamples: Final[int] = ...
    SampleFormat: Final[int] = ...
    SMinSampleValue: Final[int] = ...
    SMaxSampleValue: Final[int] = ...
    TransferRange: Final[int] = ...
    ClipPath: Final[int] = ...
    XClipPathUnits: Final[int] = ...
    YClipPathUnits: Final[int] = ...
    Indexed: Final[int] = ...
    JPEGTables: Final[int] = ...
    OPIProxy: Final[int] = ...
    JPEGProc: Final[int] = ...
    JPEGInterchangeFormat: Final[int] = ...
    JPEGInterchangeFormatLength: Final[int] = ...
    JPEGRestartInterval: Final[int] = ...
    JPEGLosslessPredictors: Final[int] = ...
    JPEGPointTransforms: Final[int] = ...
    JPEGQTables: Final[int] = ...
    JPEGDCTables: Final[int] = ...
    JPEGACTables: Final[int] = ...
    YCbCrCoefficients: Final[int] = ...
    YCbCrSubSampling: Final[int] = ...
    YCbCrPositioning: Final[int] = ...
    ReferenceBlackWhite: Final[int] = ...
    XMLPacket: Final[int] = ...
    Rating: Final[int] = ...
    RatingPercent: Final[int] = ...
    ImageID: Final[int] = ...
    CFARepeatPatternDim: Final[int] = ...
    CFAPattern: Final[int] = ...
    BatteryLevel: Final[int] = ...
    Copyright: Final[int] = ...
    ExposureTime: Final[int] = ...
    ImageResources: Final[int] = ...
    ExifTag: Final[int] = ...
    InterColorProfile: Final[int] = ...
    GPSTag: Final[int] = ...
    Interlace: Final[int] = ...
    TimeZoneOffset: Final[int] = ...
    SelfTimerMode: Final[int] = ...
    FlashEnergy: Final[int] = ...
    SpatialFrequencyResponse: Final[int] = ...
    Noise: Final[int] = ...
    FocalPlaneXResolution: Final[int] = ...
    FocalPlaneYResolution: Final[int] = ...
    FocalPlaneResolutionUnit: Final[int] = ...
    ImageNumber: Final[int] = ...
    SecurityClassification: Final[int] = ...
    ImageHistory: Final[int] = ...
    ExposureIndex: Final[int] = ...
    TIFFEPStandardID: Final[int] = ...
    SensingMethod: Final[int] = ...
    XPTitle: Final[int] = ...
    XPComment: Final[int] = ...
    XPAuthor: Final[int] = ...
    XPKeywords: Final[int] = ...
    XPSubject: Final[int] = ...
    PrintImageMatching: Final[int] = ...
    DNGVersion: Final[int] = ...
    DNGBackwardVersion: Final[int] = ...
    UniqueCameraModel: Final[int] = ...
    LocalizedCameraModel: Final[int] = ...
    CFAPlaneColor: Final[int] = ...
    CFALayout: Final[int] = ...
    LinearizationTable: Final[int] = ...
    BlackLevelRepeatDim: Final[int] = ...
    BlackLevel: Final[int] = ...
    BlackLevelDeltaH: Final[int] = ...
    BlackLevelDeltaV: Final[int] = ...
    WhiteLevel: Final[int] = ...
    DefaultScale: Final[int] = ...
    DefaultCropOrigin: Final[int] = ...
    DefaultCropSize: Final[int] = ...
    ColorMatrix1: Final[int] = ...
    ColorMatrix2: Final[int] = ...
    CameraCalibration1: Final[int] = ...
    CameraCalibration2: Final[int] = ...
    ReductionMatrix1: Final[int] = ...
    ReductionMatrix2: Final[int] = ...
    AnalogBalance: Final[int] = ...
    AsShotNeutral: Final[int] = ...
    AsShotWhiteXY: Final[int] = ...
    BaselineExposure: Final[int] = ...
    BaselineNoise: Final[int] = ...
    BaselineSharpness: Final[int] = ...
    BayerGreenSplit: Final[int] = ...
    LinearResponseLimit: Final[int] = ...
    CameraSerialNumber: Final[int] = ...
    LensInfo: Final[int] = ...
    ChromaBlurRadius: Final[int] = ...
    AntiAliasStrength: Final[int] = ...
    ShadowScale: Final[int] = ...
    DNGPrivateData: Final[int] = ...
    MakerNoteSafety: Final[int] = ...
    CalibrationIlluminant1: Final[int] = ...
    CalibrationIlluminant2: Final[int] = ...
    BestQualityScale: Final[int] = ...
    RawDataUniqueID: Final[int] = ...
    OriginalRawFileName: Final[int] = ...
    OriginalRawFileData: Final[int] = ...
    ActiveArea: Final[int] = ...
    MaskedAreas: Final[int] = ...
    AsShotICCProfile: Final[int] = ...
    AsShotPreProfileMatrix: Final[int] = ...
    CurrentICCProfile: Final[int] = ...
    CurrentPreProfileMatrix: Final[int] = ...
    ColorimetricReference: Final[int] = ...
    CameraCalibrationSignature: Final[int] = ...
    ProfileCalibrationSignature: Final[int] = ...
    AsShotProfileName: Final[int] = ...
    NoiseReductionApplied: Final[int] = ...
    ProfileName: Final[int] = ...
    ProfileHueSatMapDims: Final[int] = ...
    ProfileHueSatMapData1: Final[int] = ...
    ProfileHueSatMapData2: Final[int] = ...
    ProfileToneCurve: Final[int] = ...
    ProfileEmbedPolicy: Final[int] = ...
    ProfileCopyright: Final[int] = ...
    ForwardMatrix1: Final[int] = ...
    ForwardMatrix2: Final[int] = ...
    PreviewApplicationName: Final[int] = ...
    PreviewApplicationVersion: Final[int] = ...
    PreviewSettingsName: Final[int] = ...
    PreviewSettingsDigest: Final[int] = ...
    PreviewColorSpace: Final[int] = ...
    PreviewDateTime: Final[int] = ...
    RawImageDigest: Final[int] = ...
    OriginalRawFileDigest: Final[int] = ...
    SubTileBlockSize: Final[int] = ...
    RowInterleaveFactor: Final[int] = ...
    ProfileLookTableDims: Final[int] = ...
    ProfileLookTableData: Final[int] = ...
    OpcodeList1: Final[int] = ...
    OpcodeList2: Final[int] = ...
    OpcodeList3: Final[int] = ...
    NoiseProfile: Final[int] = ...
    ZZZTestSlong1: Final[int] = ...
    ZZZTestSlong2: Final[int] = ...
    ZZZTestSByte: Final[int] = ...
    ZZZTestSShort: Final[int] = ...
    ZZZTestDFloat: Final[int] = ...

class ExifIFD:
    """Exif tag number reference - Exif IFD"""

    ExposureTime: Final[int] = ...
    FNumber: Final[int] = ...
    ExposureProgram: Final[int] = ...
    SpectralSensitivity: Final[int] = ...
    ISOSpeedRatings: Final[int] = ...
    OECF: Final[int] = ...
    SensitivityType: Final[int] = ...
    StandardOutputSensitivity: Final[int] = ...
    RecommendedExposureIndex: Final[int] = ...
    ISOSpeed: Final[int] = ...
    ISOSpeedLatitudeyyy: Final[int] = ...
    ISOSpeedLatitudezzz: Final[int] = ...
    ExifVersion: Final[int] = ...
    DateTimeOriginal: Final[int] = ...
    DateTimeDigitized: Final[int] = ...
    OffsetTime: Final[int] = ...
    OffsetTimeOriginal: Final[int] = ...
    OffsetTimeDigitized: Final[int] = ...
    ComponentsConfiguration: Final[int] = ...
    CompressedBitsPerPixel: Final[int] = ...
    ShutterSpeedValue: Final[int] = ...
    ApertureValue: Final[int] = ...
    BrightnessValue: Final[int] = ...
    ExposureBiasValue: Final[int] = ...
    MaxApertureValue: Final[int] = ...
    SubjectDistance: Final[int] = ...
    MeteringMode: Final[int] = ...
    LightSource: Final[int] = ...
    Flash: Final[int] = ...
    FocalLength: Final[int] = ...
    Temperature: Final[int] = ...
    Humidity: Final[int] = ...
    Pressure: Final[int] = ...
    WaterDepth: Final[int] = ...
    Acceleration: Final[int] = ...
    CameraElevationAngle: Final[int] = ...
    SubjectArea: Final[int] = ...
    MakerNote: Final[int] = ...
    UserComment: Final[int] = ...
    SubSecTime: Final[int] = ...
    SubSecTimeOriginal: Final[int] = ...
    SubSecTimeDigitized: Final[int] = ...
    FlashpixVersion: Final[int] = ...
    ColorSpace: Final[int] = ...
    PixelXDimension: Final[int] = ...
    PixelYDimension: Final[int] = ...
    RelatedSoundFile: Final[int] = ...
    InteroperabilityTag: Final[int] = ...
    FlashEnergy: Final[int] = ...
    SpatialFrequencyResponse: Final[int] = ...
    FocalPlaneXResolution: Final[int] = ...
    FocalPlaneYResolution: Final[int] = ...
    FocalPlaneResolutionUnit: Final[int] = ...
    SubjectLocation: Final[int] = ...
    ExposureIndex: Final[int] = ...
    SensingMethod: Final[int] = ...
    FileSource: Final[int] = ...
    SceneType: Final[int] = ...
    CFAPattern: Final[int] = ...
    CustomRendered: Final[int] = ...
    ExposureMode: Final[int] = ...
    WhiteBalance: Final[int] = ...
    DigitalZoomRatio: Final[int] = ...
    FocalLengthIn35mmFilm: Final[int] = ...
    SceneCaptureType: Final[int] = ...
    GainControl: Final[int] = ...
    Contrast: Final[int] = ...
    Saturation: Final[int] = ...
    Sharpness: Final[int] = ...
    DeviceSettingDescription: Final[int] = ...
    SubjectDistanceRange: Final[int] = ...
    ImageUniqueID: Final[int] = ...
    CameraOwnerName: Final[int] = ...
    BodySerialNumber: Final[int] = ...
    LensSpecification: Final[int] = ...
    LensMake: Final[int] = ...
    LensModel: Final[int] = ...
    LensSerialNumber: Final[int] = ...
    Gamma: Final[int] = ...

class GPSIFD:
    """Exif tag number reference - GPS IFD"""

    GPSVersionID: Final[int] = ...
    GPSLatitudeRef: Final[int] = ...
    GPSLatitude: Final[int] = ...
    GPSLongitudeRef: Final[int] = ...
    GPSLongitude: Final[int] = ...
    GPSAltitudeRef: Final[int] = ...
    GPSAltitude: Final[int] = ...
    GPSTimeStamp: Final[int] = ...
    GPSSatellites: Final[int] = ...
    GPSStatus: Final[int] = ...
    GPSMeasureMode: Final[int] = ...
    GPSDOP: Final[int] = ...
    GPSSpeedRef: Final[int] = ...
    GPSSpeed: Final[int] = ...
    GPSTrackRef: Final[int] = ...
    GPSTrack: Final[int] = ...
    GPSImgDirectionRef: Final[int] = ...
    GPSImgDirection: Final[int] = ...
    GPSMapDatum: Final[int] = ...
    GPSDestLatitudeRef: Final[int] = ...
    GPSDestLatitude: Final[int] = ...
    GPSDestLongitudeRef: Final[int] = ...
    GPSDestLongitude: Final[int] = ...
    GPSDestBearingRef: Final[int] = ...
    GPSDestBearing: Final[int] = ...
    GPSDestDistanceRef: Final[int] = ...
    GPSDestDistance: Final[int] = ...
    GPSProcessingMethod: Final[int] = ...
    GPSAreaInformation: Final[int] = ...
    GPSDateStamp: Final[int] = ...
    GPSDifferential: Final[int] = ...
    GPSHPositioningError: Final[int] = ...

class InteropIFD:
    """Exif tag number reference - Interoperability IFD"""

    InteroperabilityIndex: Final[int] = ...



from pathlib import Path

import pytest
from typer.testing import CliRunner

from src.main import app

# Import path helpers from conftest
from tests.conftest import get_jpg_test_file, get_png_test_file, get_test_images_dir

runner = CliRunner()

# Test file paths (cross-platform)
JPG_TEST_FILE = get_jpg_test_file()
PNG_TEST_FILE = get_png_test_file()
TEST_DIR = get_test_images_dir()


# ============== Success Case Tests ==============


@pytest.mark.parametrize("x", [JPG_TEST_FILE, PNG_TEST_FILE])
def test_read_command_single_file_success(x):
    """
    Test the 'read' command with a single file (JPG and PNG).
    """
    result = runner.invoke(app, ["read", str(x)])

    assert result.exit_code == 0, f"Failed with: {result.stdout}"
    assert "Reading" in result.stdout
    assert Path(x).name in result.stdout


@pytest.mark.parametrize("ext", ["jpg", "png"])
def test_read_command_recursive_directory_success(ext):
    """
    Test the 'read' command with recursive directory processing.
    """
    result = runner.invoke(app, ["read", TEST_DIR, "-r", "-ext", ext])

    assert result.exit_code == 0, f"Failed with: {result.stdout}"
    assert "Reading" in result.stdout


def test_read_command_requires_ext_with_recursive():
    """
    Test that --recursive requires --extension flag.
    """
    result = runner.invoke(app, ["read", TEST_DIR, "-r"])

    assert result.exit_code != 0
    # Should fail with bad parameter error


def test_read_command_requires_recursive_with_ext():
    """
    Test that --extension requires --recursive flag.
    """
    result = runner.invoke(app, ["read", JPG_TEST_FILE, "-ext", "jpg"])

    assert result.exit_code != 0


# ============== Error Case Tests ==============


def test_read_command_file_not_found():
    """
    Test that the app handles missing files gracefully.
    """
    result = runner.invoke(app, ["read", "ghost_file.jpg"])

    # Typer returns exit code 2 (Usage Error) for bad arguments
    assert result.exit_code == 2
    assert "Invalid value for 'FILE_PATH'" in result.stderr
    assert "does not exist" in result.stderr



"""
Read command - Display metadata from files.

This command reads and displays metadata from image files in a
formatted table view. Supports single files or recursive directory processing.
"""

import logging
from pathlib import Path

import typer
from rich.console import Console

from src.services.metadata_factory import MetadataFactory
from src.utils.display import print_metadata_table
from src.utils.get_target_files import get_target_files

console = Console()
log = logging.getLogger("metadata-scrubber")


# fmt: off
def read(
    file_path: Path = typer.Argument(
        exists=True,  # Must exist on the filesystem
        file_okay=True,  # Can be a file
        dir_okay=True,  # Can be a directory
        readable=True,  # Must be readable (permissions check)
        resolve_path=True,  # Auto-convert to absolute path
        help="The path to the file you want to process",
    ),
    recursive: bool = typer.Option(False, "--recursive", "-r", help="Recursively process files in the specified directory."),
    ext: str = typer.Option(None,"--extension", "-ext", help="The file extension to filter by. eg: jpg, png, pdf"),
):
    if recursive and not ext:
        raise typer.BadParameter("If you provide --recursive or -r, you must also provide --extension or -ext.")
    if ext and not recursive:
        raise typer.BadParameter("If you provide --extension or -ext, you must also provide --recursive or -r.")

    for file in get_target_files(file_path, ext) if recursive else [file_path]:
        try:
            # Get the correct object from the factory
            handler = MetadataFactory.get_handler(str(file))

            # Read
            console.print(f"🔎 Reading [bold cyan]{file.name}[/bold cyan]...")
            current_data = handler.read()
            if log.isEnabledFor(logging.DEBUG):
                # if verbose mode is enabled, log the Info
                log.info(f"Successfully read metadata from {file.name}")
            print_metadata_table(current_data)

        except Exception as e:
            # display error in console
            console.print(f"❌ [bold red]Skipped[/bold red] [cyan]{file.name}[/cyan]: [dim]{e}[/dim]")

            # LOG: Full technical details (Stack trace) for you to debug

            if log.isEnabledFor(logging.DEBUG):
                # if verbose mode is enabled, log the traceback
                log.error(f"Failed to read metadata from {file}", exc_info=True)



"""
Abstract base class for metadata handlers.

This module defines the MetadataHandler ABC which establishes the interface
for all metadata handlers. Concrete implementations (ImageHandler, PDFHandler, etc.)
must implement the read, wipe, and save methods.
"""

from abc import ABC, abstractmethod
from typing import Any


class MetadataHandler(ABC):
    """
    Abstract base class for all metadata handlers.

    Defines the common interface for reading, modifying, and saving
    metadata across different file types. All concrete handlers must
    implement the abstract methods.

    Attributes:
        filepath: Path to the file being processed.
        metadata: Dictionary containing the extracted metadata.
        processed_metadata: Dictionary containing metadata after processing.
    """

    def __init__(self, filepath: str):
        """
        Initialize the metadata handler.

        Args:
            filepath: Path to the file to process.
        """
        self.filepath = filepath
        self.metadata: dict[str, Any] = {}
        self.processed_metadata: dict[str, Any] = {}

    @abstractmethod
    def read(self) -> dict[str, Any]:
        """
        Extract metadata from the file.

        Returns:
            Dict containing the extracted metadata with human-readable keys.

        Raises:
            MetadataNotFoundError: If no metadata is found in the file.
        """
        pass

    @abstractmethod
    def wipe(self) -> None:
        """
        Remove privacy-sensitive metadata from the file.

        Updates self.processed_metadata with the cleaned metadata state.

        Raises:
            MetadataProcessingError: If an error occurs during processing.
        """
        pass

    @abstractmethod
    def save(self, output_path: str | None = None) -> None:
        """
        Save the processed file with cleaned metadata.

        Args:
            output_path: Destination path for the processed file.
                        If None, uses a default location.
        """
        pass



"""
This type stub file was generated by pyright.
"""

def split_into_segments(data): # -> list[bytes]:
    """Slices JPEG meta data into a list from JPEG binary data.
    """
    ...

def read_exif_from_file(filename): # -> bytes | None:
    """Slices JPEG meta data into a list from JPEG binary data.
    """
    ...

def get_exif_seg(segments): # -> None:
    """Returns Exif from JPEG meta data list
    """
    ...

def merge_segments(segments, exif=...): # -> bytes:
    """Merges Exif with APP0 and APP1 manipulations.
    """
    ...



"""
E2E tests for the 'scrub' command.

Tests the full CLI flow for scrubbing metadata from files.
"""

from pathlib import Path

import pytest
from typer.testing import CliRunner

from src.main import app

# Import path helpers from conftest
from tests.conftest import get_jpg_test_file, get_png_test_file, get_test_images_dir

runner = CliRunner()

# Test file paths (cross-platform)
JPG_TEST_FILE = get_jpg_test_file()
PNG_TEST_FILE = get_png_test_file()
EXAMPLES_DIR = get_test_images_dir()


@pytest.fixture
def output_dir(tmp_path):
    """Create isolated output directory for each test using tmp_path."""
    output = tmp_path / "output"
    output.mkdir(parents=True, exist_ok=True)
    return output


# ============== Success Case Tests ==============


@pytest.mark.parametrize("x", [JPG_TEST_FILE, PNG_TEST_FILE])
def test_scrub_command_single_file_success(x, output_dir):
    """
    Test the 'scrub' command with a single file.
    """
    result = runner.invoke(app, ["scrub", x, "--output", str(output_dir)])

    assert result.exit_code == 0, f"Failed with: {result.stdout}"
    # Check output file was created (has processed_ prefix)
    output_file = output_dir / f"processed_{Path(x).name}"
    assert output_file.exists()


def test_scrub_command_recursive_jpg_success(output_dir):
    """
    Test the 'scrub' command with recursive directory processing for JPG.
    Uses examples folder (smaller) for faster tests.
    """
    result = runner.invoke(
        app, ["scrub", EXAMPLES_DIR, "-r", "-ext", "jpg", "--output", str(output_dir)]
    )

    assert result.exit_code == 0, f"Failed with: {result.stdout}"
    # Check at least one output file was created
    output_files = list(output_dir.glob("processed_*.jpg"))
    assert len(output_files) > 0


def test_scrub_command_dry_run(output_dir):
    """
    Test that --dry-run doesn't create files.
    """
    result = runner.invoke(
        app, ["scrub", JPG_TEST_FILE, "--output", str(output_dir), "--dry-run"]
    )

    assert result.exit_code == 0, f"Failed with: {result.stdout}"
    assert "DRY-RUN" in result.stdout
    # No files should be created in dry-run mode
    output_file = output_dir / f"processed_{Path(JPG_TEST_FILE).name}"
    assert not output_file.exists()


def test_scrub_command_with_workers(output_dir):
    """
    Test the --workers option for concurrent processing.
    """
    result = runner.invoke(
        app,
        [
            "scrub",
            EXAMPLES_DIR,
            "-r",
            "-ext",
            "jpg",
            "--output",
            str(output_dir),
            "--workers",
            "2",
        ],
    )

    assert result.exit_code == 0, f"Failed with: {result.stdout}"


# ============== Error Case Tests ==============


def test_scrub_command_file_not_found():
    """
    Test that the app handles missing files gracefully.
    """
    result = runner.invoke(app, ["scrub", "ghost_file.jpg"])

    assert result.exit_code == 2
    assert "Invalid value" in result.stderr


def test_scrub_command_requires_ext_with_recursive():
    """
    Test that --recursive requires --extension flag.
    """
    result = runner.invoke(app, ["scrub", EXAMPLES_DIR, "-r"])

    assert result.exit_code != 0


def test_scrub_command_requires_recursive_with_ext():
    """
    Test that --extension requires --recursive flag.
    """
    result = runner.invoke(app, ["scrub", JPG_TEST_FILE, "-ext", "jpg"])

    assert result.exit_code != 0



"""
This type stub file was generated by pyright.
"""

def split(data): # -> list[Any]:
    ...

def merge_chunks(chunks): # -> bytes:
    ...

def set_vp8x(chunks):
    ...

def get_file_header(chunks): # -> bytes:
    ...

def get_exif(data): # -> None:
    ...

def insert_exif_into_chunks(chunks, exif_bytes):
    ...

def insert(webp_bytes, exif_bytes): # -> bytes:
    ...

def remove(webp_bytes): # -> bytes:
    ...



"""
Factory for creating metadata handlers.

This module provides the MetadataFactory class which uses the factory pattern
to create appropriate handler instances based on file type. It enables
extensibility for supporting new file formats (PDF, Office docs, etc.).
"""

from pathlib import Path

from src.services.image_handler import ImageHandler
from src.utils.exceptions import UnsupportedFormatError


class MetadataFactory:
    """
    Factory class for creating metadata handlers.

    Uses the factory pattern to return the appropriate handler instance
    based on file extension. This design allows easy extension to support
    new file types without modifying existing code.

    Supported formats:
        - Images: .jpg, .jpeg, .png
        - Future: .pdf, .docx, .xlsx, .pptx
    """

    @staticmethod
    def get_handler(filepath: str):
        """
        Create and return the appropriate metadata handler for a file.

        Args:
            filepath: Path to the file to process.

        Returns:
            MetadataHandler: An instance of the appropriate handler subclass.

        Raises:
            UnsupportedFormatError: If no handler is defined for the file type.
            ValueError: If the path is not a valid file.
        """
        supported_extensions = ".jpg, .jpeg, .png"
        ext = Path(filepath).suffix.lower()
        if Path(filepath).is_file():
            if ext in [".jpg", ".jpeg", ".png"]:
                return ImageHandler(filepath)

            # TODO: implement other handlers
            # elif ext == ".pdf":
            #     return PDFHandler(filepath)
            # elif ext == ".xlsx":
            #     return ExcelHandler(filepath)
            # elif ext == ".pptx":
            #     return PowerPointHandler(filepath)
            else:
                raise UnsupportedFormatError(
                    f"No handler defined for {ext} files. we curently only support {supported_extensions} files."
                )
        else:
            raise ValueError(
                f"{filepath} is not a file. if you want to process a directory, use the --recursive or -r flag."
            )



"""
This type stub file was generated by pyright.
"""

from ._common import *

def insert(exif, image, new_file=...): # -> None:
    """
    py:function:: piexif.insert(exif_bytes, filename)

    Insert exif into JPEG.

    :param bytes exif_bytes: Exif as bytes
    :param str filename: JPEG
    """
    ...



from typer.testing import CliRunner

from src.main import __version__, app

# Initialize the runner
runner = CliRunner()


def test_app_version():
    """
    Ensure the --version flag prints the correct version and exits cleanly.
    """
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0, "The app crashed!"
    assert __version__ in result.stdout, "Version number not found in output"


def test_app_help():
    """
    Ensure the --help flag works.
    """
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "Metadata Scrubber Tool" in result.stdout



"""
Custom exceptions for metadata processing operations.

This module defines a hierarchy of exceptions used throughout the
metadata scrubber tool for handling various error conditions.
"""


class MetadataException(Exception):
    """Base class for all metadata-related exceptions."""

    pass


class UnsupportedFormatError(MetadataException):
    """Raised when attempting to process an unsupported file format."""

    pass


class MetadataNotFoundError(MetadataException):
    """Raised when no metadata is found in a file."""

    pass


class MetadataProcessingError(MetadataException):
    """Raised when an error occurs during metadata processing."""

    pass



"""Generate PNG test images with metadata."""

import random
from pathlib import Path

from PIL import Image, PngImagePlugin

dest_dir = Path("tests/assets/test_images")

# Generate 45 PNG images with various metadata
for i in range(1, 46):
    # Create a simple colored image
    r, g, b = random.randint(50, 200), random.randint(50, 200), random.randint(50, 200)
    img = Image.new("RGB", (100, 100), color=(r, g, b))

    # Add PNG text metadata
    pnginfo = PngImagePlugin.PngInfo()
    pnginfo.add_text("Author", f"Test Author {i}")
    pnginfo.add_text("Description", f"Test image number {i} for metadata scrubbing")
    pnginfo.add_text("Software", "Python PIL Test Generator")
    pnginfo.add_text("Copyright", f"Copyright 2024 Test {i}")
    pnginfo.add_text("Creation Time", f"2024-01-{i % 28 + 1:02d}")

    # Save with metadata
    filename = dest_dir / f"generated_test_{i:02d}.png"
    img.save(filename, pnginfo=pnginfo)

print("Generated 45 PNG images with metadata")
print(f"Total PNG count: {len(list(dest_dir.glob('*.png')))}")



"""
Batch processing service for metadata operations.

This module provides handler-agnostic batch processing that works with any
MetadataHandler subclass (images now, PDF/Office docs in future).

Supports concurrent processing via ThreadPoolExecutor for efficient handling
of large batches (1000+ files).
"""

import logging
import threading
from collections.abc import Callable, Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from pathlib import Path

from rich.console import Console

from src.services.metadata_factory import MetadataFactory

log = logging.getLogger("metadata-scrubber")
console = Console()


@dataclass
class FileResult:
    """Result of processing a single file."""

    filepath: Path
    success: bool
    action: str  # "scrubbed", "skipped", "dry-run"
    output_path: Path | None = None
    error: str | None = None


@dataclass
class BatchSummary:
    """Aggregated statistics for batch processing."""

    total: int = 0
    success: int = 0
    skipped: int = 0
    failed: int = 0
    dry_run: bool = False
    output_dir: Path | None = None
    results: list[FileResult] = field(default_factory=list)


class BatchProcessor:
    """
    Handler-agnostic batch processor for metadata operations.

    Works with any MetadataHandler subclass via MetadataFactory.
    Supports dry-run mode, automatic duplicate suffix handling,
    and concurrent processing via ThreadPoolExecutor.
    """

    def __init__(
        self,
        output_dir: str | None = None,
        dry_run: bool = False,
        max_workers: int = 4,
    ):
        """
        Initialize the batch processor.

        Args:
            output_dir: Directory to save processed files. Defaults to "./scrubbed".
            dry_run: If True, preview what would be processed without writing files.
            max_workers: Maximum number of concurrent worker threads. Defaults to 4.
        """
        self.output_dir = Path(output_dir) if output_dir else Path("./scrubbed")
        self.dry_run = dry_run
        self.max_workers = max_workers
        self.results: list[FileResult] = []

        # Thread synchronization
        self._path_lock = threading.Lock()  # Protects unique path generation
        self._results_lock = threading.Lock()  # Protects results list

    def process_file(self, file: Path) -> FileResult:
        """
        Process a single file through the read→wipe→save pipeline.

        Uses MetadataFactory to get the appropriate handler, so this method
        automatically works with any file type that has a registered handler.

        Args:
            file: Path to the file to process.

        Returns:
            FileResult with success status and details.
        """
        output_path: Path | None = None  # Track reserved path for cleanup

        try:
            # Dry-run mode: just report what would happen
            if self.dry_run:
                # Verify the file can be handled (will raise if not)
                MetadataFactory.get_handler(str(file))
                output_path = self._get_unique_output_path(file, reserve=False)
                result = FileResult(
                    filepath=file,
                    success=True,
                    action="dry-run",
                    output_path=output_path,
                )
                self._append_result(result)
                log.debug(f"[DRY-RUN] Would process: {file}")
                return result

            # Get handler from factory
            handler = MetadataFactory.get_handler(str(file))

            # Execute the read → wipe → save pipeline
            handler.read()
            handler.wipe()

            output_path = self._get_unique_output_path(file)
            handler.save(str(output_path))

            result = FileResult(
                filepath=file,
                success=True,
                action="scrubbed",
                output_path=output_path,
            )
            self._append_result(result)
            if log.isEnabledFor(logging.DEBUG):
                # if verbose mode is enabled, log the Info
                log.info(f"✅ Scrubbed: {file.name} → {output_path}")
            return result

        except Exception as e:
            # Cleanup: remove empty placeholder file if reservation failed
            self._cleanup_reserved_path(output_path)

            result = FileResult(
                filepath=file,
                success=False,
                action="skipped",
                error=str(e),
            )
            self._append_result(result)
            if log.isEnabledFor(logging.DEBUG):
                # if verbose mode is enabled, log the traceback
                log.warning(f"⚠️ Skipped {file.name}: {e}")
            return result

    def process_batch(
        self,
        files: Iterable[Path],
        progress_callback: Callable[[FileResult], None] | None = None,
    ) -> list[FileResult]:
        """
        Process multiple files concurrently using ThreadPoolExecutor.

        Args:
            files: Iterable of file paths to process.
            progress_callback: Optional callback called after each file completes.
                              Receives the FileResult for progress updates.

        Returns:
            List of FileResult objects for all processed files.
        """
        file_list = list(files)

        if not file_list:
            return self.results

        # Used ThreadPoolExecutor for I/O-bound concurrent processing
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            # Submit all files for processing
            future_to_file = {
                executor.submit(self.process_file, file): file for file in file_list
            }

            # Collect results as they complete
            for future in as_completed(future_to_file):
                result = future.result()
                if progress_callback:
                    progress_callback(result)

        return self.results

    def process_batch_sequential(self, files: Iterable[Path]) -> list[FileResult]:
        """
        Process files sequentially (legacy behavior for debugging).

        Args:
            files: Iterable of file paths to process.

        Returns:
            List of FileResult objects for all processed files.
        """
        for file in files:
            self.process_file(file)
        return self.results

    def get_summary(self) -> BatchSummary:
        """
        Return aggregated statistics for all processed files.

        Returns:
            BatchSummary with counts and result details.
        """
        with self._results_lock:
            results_copy = list(self.results)

        summary = BatchSummary(
            total=len(results_copy),
            success=sum(
                1 for r in results_copy if r.success and r.action == "scrubbed"
            ),
            skipped=sum(1 for r in results_copy if not r.success),
            dry_run=self.dry_run,
            output_dir=self.output_dir,
            results=results_copy,
        )
        # Count dry-run as separate from success for clarity
        if self.dry_run:
            summary.success = sum(1 for r in results_copy if r.success)
        return summary

    def _append_result(self, result: FileResult) -> None:
        """Thread-safe append to results list."""
        with self._results_lock:
            self.results.append(result)

    def _cleanup_reserved_path(self, output_path: Path | None) -> None:
        """
        Remove empty placeholder file created during path reservation.

        Called when processing fails after _get_unique_output_path() reserved
        a path via touch(). Only removes files that are empty (0 bytes) to
        avoid deleting partially written data.

        Args:
            output_path: Path that was reserved, or None if not yet reserved.
        """
        if output_path is None:
            return

        try:
            if output_path.exists() and output_path.stat().st_size == 0:
                output_path.unlink()
                log.debug(f"Cleaned up empty placeholder: {output_path}")
        except OSError:
            # Best effort cleanup - don't fail if we can't delete
            pass

    def _get_unique_output_path(self, file: Path, reserve: bool = True) -> Path:
        """
        Generate unique output path with suffix (_1, _2) if file exists.

        Thread-safe: uses lock to prevent race conditions during concurrent processing.

        Args:
            file: Original file path.
            reserve: If True, create placeholder file to reserve the path.
                     Set to False for dry-run mode.

        Returns:
            Unique path in output directory that doesn't conflict with existing files.
        """
        with self._path_lock:
            # creates the destination directory if it doesn't exist
            self.output_dir.mkdir(parents=True, exist_ok=True)

            base_name = file.stem
            extension = file.suffix
            output_path = self.output_dir / f"processed_{base_name}{extension}"

            # If file exists, add incrementing suffix
            counter = 1
            while output_path.exists():
                output_path = (
                    self.output_dir / f"processed_{base_name}_{counter}{extension}"
                )
                counter += 1

            # Create empty placeholder to reserve the path (skip in dry-run)
            if reserve:
                output_path.touch()

            return output_path
